package net.projectmonkey.object.mapper.construction.converter;

import net.projectmonkey.object.mapper.ObjectMappingService;
import net.projectmonkey.object.mapper.construction.PopulationContext;
import net.projectmonkey.object.mapper.construction.type.GenericTypeUtils;
import net.projectmonkey.object.mapper.context.ExecutionContext;
import net.projectmonkey.object.mapper.util.Iterables;
import net.projectmonkey.object.mapper.util.TypeResolver;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

import static net.projectmonkey.object.mapper.construction.type.GenericTypeUtils.TypeAndClass;

/*
 *
 *  * Copyright 2012 the original author or authors.
 *  *
 *  * Licensed under the Apache License, Version 2.0 (the "License");
 *  * you may not use this file except in compliance with the License.
 *  * You may obtain a copy of the License at
 *  *
 *  *      http://www.apache.org/licenses/LICENSE-2.0
 *  *
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS,
 *  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  * See the License for the specific language governing permissions and
 *  * limitations under the License.
 *
 */

/**
 * Converts {@link java.util.Collection} and array instances to {@link java.util.Collection} instances.
 *
 * N.B. This implementation ALWAYS replaces the destination collection
 * with the converted contents of the source.
 *
 * An alternative strategy would be to attempt to merge the destination
 * objects based on position in the collection and remove any superfluous
 * destinations but this seemed likely to be error prone.
 *
 * <p/>
 * Adapted from ModelMapper (modelmapper.org)
 *
 * @author Jonathan Halterman
 * @author Andy Moody
 */
public class CollectionConverter extends IterableConverter<Object, Collection<Object>>
{

	public static final CollectionConverter INSTANCE = new CollectionConverter();

	private CollectionConverter(){}

	@Override
	public boolean canConvert(final PopulationContext<?,?> context)
	{
		Object source = context.getSource();
		return source != null && Collection.class.isAssignableFrom(context.getDestinationType()) && Iterables.isIterable(source.getClass());
	}

	@Override
	protected Collection<Object> createDestination(final Class<Collection<Object>> destinationType, final int length, final TypeAndClass elementType)
	{
		Collection<Object> toReturn = null;
		if(destinationType.isInterface())
		{
			if(Set.class.isAssignableFrom(destinationType))
			{
				toReturn = new LinkedHashSet<Object>(length);
			}
			else
			{
				toReturn = new ArrayList<Object>(length);
			}
		}
		else
		{
			toReturn = getMappingService().instantiate(destinationType);
		}
		return toReturn;
	}

	@Override
	protected TypeAndClass getElementType(final Class<?> destinationType, final PopulationContext<Object, Collection<Object>> context)
	{
		TypeAndClass elementType = GenericTypeUtils.INSTANCE.getArgumentType(destinationType, context);
		return elementType.getClazz() == TypeResolver.Unknown.class ? new TypeAndClass(Object.class) : elementType;
	}

	@Override
	protected void setElement(final Collection<Object> destination, final Object element, int index)
	{
		destination.add(element);
	}

	public ObjectMappingService getMappingService()
	{
		return ExecutionContext.getMappingService();
	}


}
